from dataclasses import dataclass
from typing import List, Any


@dataclass
class Command:

    command: str
    arguments: List[Any] = None
    stdout: str = None
    stderr: str = None
    stdin: str = None

    @classmethod
    def from_spec(cls, spec):
        return cls(**spec)


@dataclass
class CommandArgument:
    """
    Attributes
    ----------
    parts : list of any of str, File, Parameter

    """

    parts: List[Any]



import datetime
import enum

import zarr

from hpcflow.utils import make_workflow_id, get_time_stamp
from hpcflow.config import Config


class WorkflowInteraction(enum.Enum):

    CREATE = 0


class Workflow:
    def __init__(self, tasks):
        self.tasks = tasks


class Task:
    def __init__(self, schema, parameter_values):
        self.schema = schema
        self.parameter_values


class TaskSchema:
    def __init__(self, parameters):
        self.parameters = parameters


def make_workflow():

    Config.set_config()

    workflow = Workflow(
        tasks=[
            Task(
                schema=TaskSchema(parameters=("A", "B")),
                parameter_values={"A": 1, "B": 2},
            ),
            Task(schema=TaskSchema(parameters=("D")), parameter_values={"D": 4}),
        ]
    )

    id_ = "jd3ifk"  # make_workflow_id()   TEMP
    store = zarr.DirectoryStore(f"workflow_{id_}.zarr")
    root = zarr.group(store=store, overwrite=True)
    parameter_group = root.create_group("parameters")
    history_group = root.create_group("history")

    ts_str = get_time_stamp()

    make_history = history_group.create_group(ts_str)
    make_history.attrs["timestamp"] = ts_str
    make_history.attrs["interaction"] = WorkflowInteraction.CREATE.name



from dataclasses import dataclass
from operator import itemgetter
from typing import List, Optional
from hpcflow.loop import Loop

from hpcflow.object_list import TaskList
from hpcflow.parameters import (
    InputSource,
    ParameterPropagationMode,
    SchemaInput,
    ValueSequence,
)
from hpcflow.task import Task, TaskTemplate
from hpcflow.utils import (
    get_in_container,
    get_relative_path,
    group_by_dict_key_values,
    set_in_container,
)


class WorkflowTemplate:
    def __init__(
        self,
        task_templates: Optional[List[TaskTemplate]] = None,
        loops: Optional[List[Loop]] = None,
    ):

        self.parameter_data = []
        self.parameter_mapping = []
        self.elements = []
        self.tasks = TaskList()
        self.element_indices = []
        self.name_repeat_indices = []

        for task_template in task_templates or []:
            self.add_task(task_template)

    def get_possible_input_sources(
        self, schema_input: SchemaInput, new_task: TaskTemplate, new_index: int
    ):
        """Enumerate the possible sources for an input of a new task, given a proposed
        placement of that task."""

        # Get parameters provided by tasks up to `new_index`:
        task_sources = {}
        for task in self.tasks[:new_index]:
            provided = tuple(
                i
                for i in task.template.provides_parameters
                if i.typ == schema_input.typ
            )
            if provided:
                task_sources.update({(task.index, task.unique_name): provided})

        out = {
            "imports": {},
            "tasks": task_sources,
            "has_local": schema_input.typ in new_task.defined_input_types,
            # TODO: there *might* be local definition of a parameter in the form of the input files/writers specified?
            "has_default": schema_input.default_value is not None,
        }
        return out

    def ensure_input_sources(self, new_task: TaskTemplate, new_index: int):
        """Check valid input sources are specified for a new task to be added to the
        workflow in a given position. If none are specified, set them according to the
        default behaviour."""

        all_sources = {}
        for schema_input in new_task.all_schema_inputs:

            all_sources.update(
                {
                    schema_input.typ: self.get_possible_input_sources(
                        schema_input, new_task, new_index
                    )
                }
            )

            # if sources are specified, check they are resolvable:
            for specified_source in new_task.input_sources.get(schema_input.typ, []):
                specified_source.validate(schema_input, new_task, self)

        print(f"\nall_sources: {all_sources}")
        print(f"new_task.unsourced_inputs: {new_task.unsourced_inputs}")

        # if an input is not specified at all in the `inputs` dict (what about when list?),
        # then check if there is an input files entry for associated inputs,
        # if there is

        # set source for any unsourced inputs:
        for input_type in new_task.unsourced_inputs:
            inp_i_sources = all_sources[input_type]

            print(f"sources: {inp_i_sources}")

            # input may not be required

            # set source for this input:
            if inp_i_sources["has_local"]:
                new_sources = [InputSource("local")]

            elif inp_i_sources["tasks"]:
                # we can only take parameters with implicit propagation mode:
                params_info = []
                for (task_idx, task_name), params in inp_i_sources["tasks"].items():
                    for i in params:
                        if i.propagation_mode == ParameterPropagationMode.IMPLICIT:
                            params_info.append((i.input_or_output, task_idx, task_name))

                # sort by output/input (output first), then by task index (highest first):
                params_info = sorted(
                    params_info, key=itemgetter(0, 1), reverse=True
                )  # TODO: TEST THIS!!!
                new_sources = [
                    InputSource(f"tasks.{params_info[0][2]}.{params_info[0][0]}s")
                ]

            else:
                # input may not need defining (if all associated input files are passed
                # and the input does not appear in any action commands)
                new_sources = None

            new_task.input_sources.update({input_type: new_sources})

    def add_task(self, task_template: TaskTemplate):

        # TODO: can't do this check yet because required inputs of different elements may be different
        # e.g. if an input file is passed for some elements but not others.

        self.ensure_input_sources(
            task_template, len(self.tasks)
        )  # modifies task_template.input_sources
        # at this point the source for each input should be decided and well-defined.

        element_indices = []

        add_sequences = [  # treat the base inputs and resources as single-item sequences:
            ValueSequence(
                path=["inputs"],
                values=[
                    {
                        param.parameter.typ: param.value
                        for param in task_template.get_non_sub_parameter_input_values()
                    }
                ],
                nesting_order=-1,
            ),
            ValueSequence(
                path=["resources"], values=[task_template.resources], nesting_order=-1
            ),
        ]  # treat sub-parameter input values as single-item sequences
        for i in task_template.get_sub_parameter_input_values():
            add_sequences.append(
                ValueSequence(
                    path=["inputs", i.parameter.typ] + i.path,
                    values=[i.value],
                    nesting_order=-1,
                )
            )

        multi = []
        input_map_indices = {}

        sequences = add_sequences + task_template.sequences
        for i in sequences:
            # add each sequence data:
            next_param_idx = len(self.parameter_data)
            num_values = len(i.values)
            self.parameter_data.extend([{"is_set": True, "data": j} for j in i.values])
            self.parameter_mapping.append(
                list(range(next_param_idx, next_param_idx + num_values))
            )
            param_map_idx = len(self.parameter_mapping) - 1
            input_map_indices[tuple(i.path)] = param_map_idx
            nesting_order_i = (
                task_template.nesting_order[tuple(i.path)] if num_values > 1 else -1
            )
            multi.append(
                {
                    "multiplicity": num_values,
                    "nesting_order": nesting_order_i,
                    "address": i.path,
                }
            )

        init_multi = WorkflowTemplate.resolve_initial_elements(multi)
        output_map_indices = {}
        num_elems = len(init_multi)
        for schema in task_template.schemas:
            for output in schema.outputs:
                next_dat_idx = len(self.parameter_data)
                next_map_idx = len(self.parameter_mapping)
                out_data = [{"is_set": False, "data": None} for _ in range(num_elems)]
                out_param_map = list(range(next_dat_idx, next_dat_idx + num_elems))
                self.parameter_data.extend(out_data)
                self.parameter_mapping.append(out_param_map)
                output_map_indices[output.typ] = next_map_idx

        for i_idx, i in enumerate(init_multi):
            element_indices.append(len(self.elements))
            self.elements.append(
                {
                    "inputs": [
                        {
                            "path": k,
                            "parameter_mapping_index": input_map_indices[tuple(k)],
                            "data_index": v,
                        }
                        for k, v in i["value_index"].items()
                    ],
                    "outputs": [
                        {
                            "path": ("outputs", k),
                            "parameter_mapping_index": v,
                            "data_index": i_idx,
                        }
                        for k, v in output_map_indices.items()
                    ],
                }
            )

        self.element_indices.append(element_indices)
        self.name_repeat_indices.append(
            sum(i.template.name == task_template.name for i in self.tasks) + 1
        )
        task = Task(task_template, self, len(self.tasks))
        self.tasks.add_object(task)

    @staticmethod
    def resolve_initial_elements(multi):
        """
        Parameters
        ----------
        multi : list of dict
            Each list item represents a sequence of values with keys:
                multiplicity: int
                nesting_order: int
                address : str
        """

        # order by nesting order (so lower nesting orders will be fastest-varying):
        multi_srt = sorted(multi, key=lambda x: x["nesting_order"])
        multi_srt_grp = group_by_dict_key_values(multi_srt, "nesting_order")

        elements = [{"value_index": {}}]
        for para_sequences in multi_srt_grp:

            # check all equivalent nesting_orders have equivalent multiplicities
            all_multis = {i["multiplicity"] for i in para_sequences}
            if len(all_multis) > 1:
                raise ValueError(
                    f"All sequences with the same `nesting_order` must have the same "
                    f"multiplicity, but found multiplicities {list(all_multis)!r} for "
                    f"`nesting_order` of {para_sequences[0]['nesting_order']}."
                )

            new_elements = []
            for val_idx in range(para_sequences[0]["multiplicity"]):
                for element in elements:
                    new_elements.append(
                        {
                            "value_index": {
                                **element["value_index"],
                                **{i["address"]: val_idx for i in para_sequences},
                            }
                        }
                    )
            elements = new_elements

        return elements

    def add_task_after(self, task):
        pass

    def add_task_before(self, task):
        pass

    def remove_task(self, task):
        pass

    def get_input_values(self, task_index, parameter_path):
        """Get the value of an input for each element in a task."""
        return [
            self.get_input_value(
                task_index=task_index, element_index=i, parameter_path=parameter_path
            )
            for i in range(self.tasks[task_index].num_elements)
        ]

    def get_input_value(self, task_index, element_index, parameter_path):

        element = self.elements[self.tasks[task_index].element_indices[element_index]]
        current_value = None
        for input_i in element["inputs"]:

            param_data_idx = self.parameter_mapping[input_i["parameter_mapping_index"]][
                input_i["data_index"]
            ]

            is_parent = False
            is_update = False
            try:
                rel_path_parts = get_relative_path(parameter_path, input_i["path"])
                is_parent = True
            except ValueError:
                try:
                    update_path = get_relative_path(input_i["path"], parameter_path)
                    is_update = True
                except ValueError:
                    pass

            if is_parent:
                # replace current value:
                final_data_path = (param_data_idx, "data", *rel_path_parts)
                try:
                    current_value = get_in_container(
                        self.parameter_data, final_data_path
                    )  # or use Zarr to get from persistent
                except TypeError:
                    # import traceback

                    # traceback.print_exc()
                    pass

            elif is_update:
                # update sub-part of current value
                update_data = self.parameter_data[param_data_idx][
                    "data"
                ]  # or use Zarr to get from persistent
                set_in_container(current_value, update_path, update_data)

        return current_value

    def make_workflow(self, path):
        # TODO: make the workflow and save to path
        wk = Workflow(path)
        return wk

    @classmethod
    def from_spec(cls, spec, all_schemas, all_parameters):

        # initialise task templates:
        tasks = []
        for i in spec.pop("tasks"):
            tasks.append(TaskTemplate.from_spec(i, all_schemas, all_parameters))
        spec["task_templates"] = tasks

        return cls(**spec)


@dataclass
class Workflow:
    tasks: List[Task]

    def rename(self, new_name):
        pass

    def add_submission(self, filter):
        pass


@dataclass
class WorkflowBlueprint:
    """For pre-built workflows that are simpler to parametrise (e.g. fitting workflows)."""

    workflow_template: WorkflowTemplate



from dataclasses import dataclass

import click

from hpcflow.cli import cli


@dataclass
class HPCFlow:
    """Class for instantiating an HPCFlow application, which may provide, for instance,
    custom parameter logic, over the top of that provided by hpcflow."""

    name: str
    version: str

    def __post_init__(self):
        self.CLI = self.make_CLI()

    def make_CLI(self):
        new_CLI = click.Group(name=self.name)
        new_CLI = click.version_option(package_name=self.name, version=self.version)(
            new_CLI
        )
        # add hpcflow CLI as a sub command:
        new_CLI.add_command(cli)
        for name, command in cli.commands.items():
            # add each hpcflow command as a new CLI command:
            new_CLI.add_command(command, name=name)
        return new_CLI



class DotAccessObjectList:
    def __init__(self, *objects, access_attribute, descriptor):
        self._objects = list(objects)
        self._access_attribute = access_attribute
        self._descriptor = descriptor

        if self._objects:

            if not hasattr(self._objects[0], self._access_attribute):
                raise TypeError(
                    f"Objects do not have attribute {self._access_attribute!r}."
                )

    def __len__(self):
        return len(self._objects)

    def __repr__(self):
        return repr(self._objects)

    def __str__(self):
        return str(self._objects)

    def __getitem__(self, key):
        return self._objects.__getitem__(key)

    def __iter__(self):
        return self._objects.__iter__()

    def __contains__(self, item):
        return self._objects.__contains__(item)

    def __eq__(self, other):
        return self._objects == other

    def __getattr__(self, attribute):
        for obj in self._objects:
            if getattr(obj, self._access_attribute) == attribute:
                return obj

        obj_list_fmt = ", ".join(
            [f'"{getattr(i, self._access_attribute)}"' for i in self._objects]
        )
        msg = (
            f"{self._descriptor.title()} {attribute!r} does not exist. Available "
            f"{self._descriptor}s are: {obj_list_fmt}."
        )
        raise AttributeError(msg)

    def __dir__(self):
        return super().__dir__() + [
            getattr(i, self._access_attribute) for i in self._objects
        ]

    def add_object(self, obj, index=-1):
        if not hasattr(obj, self._access_attribute):
            raise TypeError(
                f"Object does not have attribute {self._access_attribute!r}."
            )
        if index < 0:
            index += len(self) + 1
        self._objects = self._objects[:index] + [obj] + self._objects[index:]


class TaskList(DotAccessObjectList):
    """A list-like container for a task-like list with dot-notation access by task
    unique-name."""

    def __init__(self, *task_likes):
        super().__init__(
            *task_likes, access_attribute="unique_name", descriptor="task-like"
        )


class GroupList(DotAccessObjectList):
    """A list-like container for the task schema group list with dot-notation access by
    group name."""

    def __init__(self, *groups):
        super().__init__(*groups, access_attribute="name", descriptor="group")


class ExecutablesList(DotAccessObjectList):
    """A list-like container for environment executables with dot-notation access by
    executable label."""

    def __init__(self, *execs):
        super().__init__(*execs, access_attribute="label", descriptor="executable")


def index(obj_lst, obj):
    for idx, i in enumerate(obj_lst._objects):
        if obj is i:
            return idx
    raise ValueError(f"{obj!r} not in list.")



import keyword
from pathlib import Path
import random
import re
import string
from datetime import datetime, timezone
from typing import Mapping

from hpcflow.errors import InvalidIdentifier


def make_workflow_id():
    length = 12
    chars = string.ascii_letters + "0123456789"
    return "".join(random.choices(chars, k=length))


def get_time_stamp():
    return datetime.now(timezone.utc).astimezone().strftime("%Y.%m.%d_%H:%M:%S_%z")


def get_duplicate_items(lst):
    """Get a list of all items in an iterable that appear more than once, assuming items
    are hashable.

    Examples
    --------
    >>> get_duplicate_items([1, 1, 2, 3])
    [1]

    >>> get_duplicate_items([1, 2, 3])
    []

    >>> get_duplicate_items([1, 2, 3, 3, 3, 2])
    [2, 3, 2]

    """
    seen = []
    return list(set(x for x in lst if x in seen or seen.append(x)))


def check_valid_py_identifier(name):
    """Check a string is (roughly) a valid Python variable identifier and if so return it
    in lower-case.

    Notes
    -----
    Will be used for:
      - task objective name
      - task method
      - task implementation
      - parameter type
      - parameter name
      - loop name
      - element group name

    """
    trial_name = name[1:].replace("_", "")  # "internal" underscores are allowed
    if (
        not name
        or not (name[0].isalpha() and ((trial_name[1:] or "a").isalnum()))
        or keyword.iskeyword(name)
        or name == "add_object"  # method of `DotAccessObjectList`
    ):
        raise InvalidIdentifier(f"Invalid string for identifier: {name!r}")

    return name.lower()


def group_by_dict_key_values(lst, *keys):
    """Group a list of dicts according to specified equivalent key-values.

    Parameters
    ----------
    lst : list of dict
        The list of dicts to group together.
    keys : tuple
        Dicts that have identical values for all of these keys will be grouped together
        into a sub-list.

    Returns
    -------
    grouped : list of list of dict

    Examples
    --------
    >>> group_by_dict_key_values([{'a': 1}, {'a': 2}, {'a': 1}], 'a')
    [[{'a': 1}, {'a': 1}], [{'a': 2}]]

    """

    grouped = [[lst[0]]]
    for lst_item in lst[1:]:
        for group_idx, group in enumerate(grouped):

            try:
                is_vals_equal = all(lst_item[k] == group[0][k] for k in keys)

            except KeyError:
                # dicts that do not have all `keys` will be in their own group:
                is_vals_equal = False

            if is_vals_equal:
                grouped[group_idx].append(lst_item)
                break

        if not is_vals_equal:
            grouped.append([lst_item])

    return grouped


def get_in_container(cont, path):
    cur_data = cont
    for idx, path_comp in enumerate(path):
        if isinstance(cur_data, (list, tuple)):
            if not isinstance(path_comp, int):
                raise TypeError(
                    f"Path component {path_comp!r} must be an integer index "
                    f"since data is a sequence: {cur_data!r}."
                )
            cur_data = cur_data[path_comp]
        elif isinstance(cur_data, Mapping):
            cur_data = cur_data[path_comp]
        else:
            raise ValueError(
                f"Data at path {path[:idx]} is not a sequence, but is of type "
                f"{type(cur_data)!r} and so sub-data cannot be extracted."
            )
    return cur_data


def set_in_container(cont, path, value):
    sub_data = get_in_container(cont, path[:-1])
    sub_data[path[-1]] = value


def get_relative_path(path1, path2):
    """Get relative path components between two paths.

    Parameters
    ----------
    path1 : tuple of (str or int or float) of length N
    path2 : tuple of (str or int or float) of length less than or equal to N

    Returns
    -------
    relative_path : tuple of (str or int or float)
        The path components in `path1` that are not in `path2`.

    Raises
    ------
    ValueError
        If the two paths do not share a common ancestor of path components, or if `path2`
        is longer than `path1`.

    Notes
    -----
    This function behaves like a simplified `PurePath(*path1).relative_to(PurePath(*path2))`
    from the `pathlib` module, but where path components can include non-strings.

    Examples
    --------
    >>> get_relative_path(('A', 'B', 'C'), ('A',))
    ('B', 'C')

    >>> get_relative_path(('A', 'B'), ('A', 'B'))
    ()

    """

    len_path2 = len(path2)
    msg = f"{path1!r} is not in the subpath of {path2!r}."

    if len(path1) < len_path2:
        raise ValueError(msg)

    for i, j in zip(path1[:len_path2], path2):
        if i != j:
            raise ValueError(msg)

    return path1[len_path2:]


def search_dir_files_by_regex(pattern, group=0, directory="."):
    vals = []
    for i in Path(directory).iterdir():
        match = re.search(pattern, i.name)
        if match:
            match_groups = match.groups()
            if match_groups:
                match = match_groups[group]
                vals.append(match)
    return vals


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)



import copy
from dataclasses import dataclass, field
import enum
from lib2to3.pytree import Base
from re import I
from typing import Dict, List, Optional, Tuple
from numpy import isin, sort

from valida.conditions import ConditionLike

from hpcflow.command_files import InputFileGenerator, OutputFileParser
from hpcflow.commands import Command, CommandArgument
from hpcflow.environment import Environment
from hpcflow.errors import MissingActionEnvironment, MissingCompatibleActionEnvironment
from hpcflow.parameters import SchemaParameter
from hpcflow.utils import classproperty


class ActionScopeType(enum.Enum):

    ALL = 0
    MAIN = 1
    PROCESSING = 2
    INPUT_FILE_GENERATOR = 3
    OUTPUT_FILE_PARSER = 4


@dataclass
class ActionScope:
    """Class to represent the identification of a subset of task schema actions by a
    filtering process.
    """

    typ: ActionScopeType
    kwargs: Optional[Dict] = field(default_factory=lambda: {})

    @classmethod
    def main(cls):
        return cls(typ=ActionScopeType.MAIN)

    @classmethod
    def processing(cls):
        return cls(typ=ActionScopeType.PROCESSING)

    @classmethod
    def input_file_generator(cls, file=None):
        return cls(typ=ActionScopeType.INPUT_FILE_GENERATOR, kwargs={"file": file})

    @classmethod
    def output_file_parser(cls, output=None):
        return cls(typ=ActionScopeType.OUTPUT_FILE_PARSER, kwargs={"output": output})


@dataclass
class ActionEnvironment:
    environment: Environment
    scope: ActionScope

    @classmethod
    def from_spec(cls, scope, env, all_envs):
        env = [i for i in all_envs if i.name == env][0]
        scope = getattr(ActionScope, scope)()
        return cls(scope=scope, environment=env)


@dataclass
class ActionCondition:
    """Class to represent a condition that must be met if an action is to be included."""

    path: List[str]
    condition: Optional[ConditionLike] = None


class Action:
    """"""

    def __init__(
        self,
        commands: List[Command],
        environments: List[ActionEnvironment],
        input_file_generators: Optional[List[InputFileGenerator]] = None,
        output_file_parsers: Optional[List[OutputFileParser]] = None,
        conditions: Optional[List[ActionCondition]] = None,
    ):
        self.commands = commands
        self.environments = environments
        self.input_file_generators = input_file_generators or []
        self.output_file_parsers = output_file_parsers or []
        self.conditions = conditions or []

    @classmethod
    def from_spec(cls, spec, all_envs, parameters, cmd_files):
        """Parse an Action definition from a JSON-like dict.

        Parameters
        ----------
        spec : dict with keys:
            commands
            input_file_generators
            output_file_parsers
            environments: dict
            environment: str
                Instead of `environments` (plural), this can be used to specify a single
                environment to be used for all resolved actions of this action.


        """
        spec = copy.deepcopy(spec)
        spec["commands"] = [Command.from_spec(i) for i in spec.get("commands", [])]
        spec["input_file_generators"] = [
            InputFileGenerator.from_spec(label, info, parameters, cmd_files)
            for label, info in spec.pop("input_files", {}).items()
        ]
        spec["output_file_parsers"] = [
            OutputFileParser.from_spec(param_typ, info, parameters, cmd_files)
            for param_typ, info in spec.pop("outputs", {}).items()
        ]
        given_envs = spec.pop("environments") or spec.pop("environment")
        if isinstance(given_envs, str):
            given_envs = {"all": given_envs}
        elif not isinstance(given_envs, dict):
            raise ValueError()

        spec["environments"] = [
            ActionEnvironment.from_spec(scope, env, all_envs)
            for scope, env in given_envs.items()
        ]
        return cls(**spec)

    def get_parameter_dependence(self, parameter: SchemaParameter):
        """Find if/where a given parameter is used by the action."""
        writer_files = [
            i.input_file
            for i in self.input_file_generators
            if parameter.parameter in i.inputs
        ]  # names of input files whose generation requires this parameter
        commands = []  # TODO: indices of commands in which this parameter appears
        out = {"input_file_writers": writer_files, "commands": commands}
        return out

    def get_resolved_action_env(
        self,
        relevant_scopes: Tuple[ActionScopeType],
        input_file_generator: InputFileGenerator = None,
        output_file_parser: OutputFileParser = None,
        commands: List[Command] = None,
    ):
        print(f"relevant_scopes: {relevant_scopes}")
        print(f"self.environments: {self.environments}")
        print(
            f"[i.scope.typ for i in self.environments]: {[i.scope.typ for i in self.environments]}"
        )
        possible = [
            i.scope.type for i in self.environments if i.scope.typ in relevant_scopes
        ]
        print(f"possible: {possible}")
        if not possible:
            if input_file_generator:
                msg = f"input file generator {input_file_generator!r}."
            elif output_file_parser:
                msg = f"output file parser {output_file_parser!r}"
            else:
                msg = f"commands {commands!r}"
            raise MissingCompatibleActionEnvironment(
                f"No compatible environment is specified for the {msg}."
            )

        # sort by scope specificity:
        possible_srt = sorted(possible, key=lambda i: i.scope.value, reverse=True)
        return possible_srt[0]

    def get_input_file_generator_action_env(
        self, input_file_generator: InputFileGenerator
    ):
        return self.get_resolved_action_env(
            relevant_scopes=(
                ActionScopeType.ALL,
                ActionScopeType.PROCESSING,
                ActionScopeType.INPUT_FILE_GENERATOR,
            ),
            input_file_generator=input_file_generator,
        )

    def get_output_file_parser_action_env(self, output_file_parser: OutputFileParser):
        return self.get_resolved_action_env(
            relevant_scopes=(
                ActionScopeType.ALL,
                ActionScopeType.PROCESSING,
                ActionScopeType.OUTPUT_FILE_PARSER,
            ),
            output_file_parser=output_file_parser,
        )

    def get_commands_action_env(self):
        return self.get_resolved_action_env(
            relevant_scopes=(ActionScopeType.ALL, ActionScopeType.MAIN),
            commands=self.commands,
        )

    def resolve_actions(self):

        cmd_acts = []
        for i in self.input_file_generators:
            act_i = InputFileGeneratorAction(
                input_file_generator=i,
                conditions=self.conditions,
                environment=self.get_input_file_generator_action_env(i),
            )
            cmd_acts.append(act_i)

        cmd_acts.append(
            CommandsAction(
                commands=self.commands,
                environment=self.get_commands_action_env(),
                conditions=self.conditions,
            )
        )

        for i in self.output_file_parsers:
            act_i = OutputFileParserAction(
                output_file_parser=i,
                environment=self.get_output_file_parser_action_env(i),
                conditions=self.conditions,
            )
            cmd_acts.append(act_i)

        return cmd_acts


@dataclass
class ResolvedAction:
    """"""

    environment: Environment
    conditions: List[ActionCondition]

    def __post_init__(self):
        # select correct environment
        pass


@dataclass
class CommandsAction(ResolvedAction):
    """Represents an action without any associated input file generators and output
    parsers."""

    commands: List[Command]


@dataclass
class InputFileGeneratorAction(ResolvedAction):
    input_file_generator: InputFileGenerator

    def __post_init__(self):
        self.conditions = (
            []
        )  # TODO: add a condition, according to non-presence of input file?


@dataclass
class OutputFileParserAction(ResolvedAction):
    output_file_parser: OutputFileParser



"""Module containing a Valida schema for the workflow spec file."""

from valida import Schema, Rule, Value, Key, Index
from valida.datapath import ListValue, MapValue

# TODO: write this schema in yaml

_RULES = []

SPEC_SCHEMA = Schema(_RULES)



from dataclasses import dataclass
from typing import List, Optional

from valida.conditions import ConditionLike

from hpcflow.typing_stubs import Parameter, ParameterPath, Task
from hpcflow.utils import check_valid_py_identifier


@dataclass
class Element:
    task: Task
    inputs: List[Parameter]
    outputs: List[Parameter]


@dataclass
class ElementFilter:

    parameter_path: ParameterPath
    condition: ConditionLike

    @classmethod
    def from_spec(cls, spec):
        raise NotImplementedError("ain't dun it yet")


@dataclass
class ElementGroup:

    name: str
    where: Optional[ElementFilter] = None
    group_by_distinct: Optional[ParameterPath] = None

    def __post_init__(self):
        self.name = check_valid_py_identifier(self.name)


@dataclass
class ElementRepeats:

    number: int
    where: Optional[ElementFilter] = None



from calendar import c


class SubParameter:
    pass


class Workflow:
    pass


class Element:
    pass


class TaskTemplate:
    pass


class TaskSchema:
    pass


class WorkflowTemplate:
    pass


class InputValue:
    pass


class Parameter:
    pass


class ParameterPath:
    pass


class Task:
    pass



"""Module containing logic for parsing workflow spec files/strings."""

from importlib import resources
from pathlib import Path

from ruamel.yaml import YAML
from valida import Schema
from hpcflow.command_files import FileSpec

from hpcflow.errors import (
    TaskSchemaSpecValidationError,
    WorkflowSpecValidationError,
    EnvironmentSpecValidationError,
)
from hpcflow.parameters import Parameter, SchemaInput, SchemaOutput
from hpcflow.task_schema import TaskSchema
from hpcflow.workflow import WorkflowTemplate
from hpcflow.environment import Executable, ExecutableInstance, Environment


def get_workflow_spec_schema():
    with resources.open_text("hpcflow.data", "workflow_spec_schema.yaml") as fh:
        schema_dat = fh.read()
    schema = Schema.from_yaml(schema_dat)
    return schema


def get_task_schema_spec_schema():
    with resources.open_text("hpcflow.data", "task_schema_spec_schema.yaml") as fh:
        schema_dat = fh.read()
    schema = Schema.from_yaml(schema_dat)
    return schema


def get_environment_spec_schema():
    with resources.open_text("hpcflow.data", "environments_spec_schema.yaml") as fh:
        schema_dat = fh.read()
    schema = Schema.from_yaml(schema_dat)
    return schema


def get_task_schemas_and_parameters():
    with resources.open_text("hpcflow.data", "task_schemas.yaml") as fh:
        yaml_str = fh.read()

    yaml = YAML(typ="safe")
    task_schemas_dat = yaml.load(yaml_str)

    task_schemas_spec_schema = get_task_schema_spec_schema()
    validated = task_schemas_spec_schema.validate(task_schemas_dat)
    if not validated.is_valid:
        raise TaskSchemaSpecValidationError(validated.get_failures_string())

    parameters = {}
    for i in task_schemas_dat["parameters"]:
        parameters.update({i["type"]: Parameter.from_spec(i)})

    cmd_files = []
    for i in task_schemas_dat["command_files"]:
        cmd_files.append(FileSpec.from_spec(i))

    envs = get_environments()

    task_schemas = {}
    for i in task_schemas_dat["task_schemas"]:
        key = (i["objective"], i.get("method"), i.get("implementation"))
        schema = TaskSchema.from_spec(i, parameters, envs, cmd_files)
        task_schemas.update({key: schema})

    return task_schemas, parameters, envs, cmd_files


def get_environments():
    with resources.open_text("hpcflow.data", "environments.yaml") as fh:
        yaml_str = fh.read()

    yaml = YAML(typ="safe")
    envs_dat = yaml.load(yaml_str)

    env_spec_schema = get_environment_spec_schema()
    validated = env_spec_schema.validate(envs_dat)
    if not validated.is_valid:
        raise EnvironmentSpecValidationError(validated.get_failures_string())

    envs = [Environment.from_spec(env_spec) for env_spec in envs_dat]
    return envs


def parse_YAML_spec_file(yaml_file):
    """Generate a WorkflowTemplate from a YAML string."""
    with Path(yaml_file).open("r") as fh:
        yaml_str = fh.read()
    return parse_YAML_spec_str(yaml_str)


def parse_YAML_spec_str(yaml_str):
    """Generate a WorkflowTemplate from a YAML string."""
    yaml = YAML(typ="safe")
    workflow_dat = yaml.load(yaml_str)
    validated = get_workflow_spec_schema().validate(workflow_dat)

    if not validated.is_valid:
        raise WorkflowSpecValidationError(validated.get_failures_string())

    task_schemas, parameters, envs, cmd_files = get_task_schemas_and_parameters()

    workflow = WorkflowTemplate.from_spec(workflow_dat, task_schemas, parameters)

    return workflow



from dataclasses import dataclass
from optparse import Option
from typing import Dict, List, Optional, Tuple, Union
from hpcflow.command_files import FileSpec

from hpcflow.element import Element, ElementFilter, ElementGroup
from hpcflow.errors import (
    MissingInputs,
    TaskTemplateInvalidNesting,
    TaskTemplateMultipleInputValues,
    TaskTemplateMultipleSchemaObjectives,
    TaskTemplateUnexpectedInput,
)
from hpcflow.object_list import GroupList, index
from hpcflow.parameters import (
    InputSource,
    InputValue,
    ParameterPath,
    SchemaInput,
    SchemaOutput,
    ValuePerturbation,
    ValueSequence,
)
from hpcflow.task_schema import TaskSchema
from hpcflow.typing_stubs import Workflow, WorkflowTemplate
from hpcflow.utils import get_duplicate_items


class TaskTemplate:
    """Parametrisation of an isolated task for which a subset of input values are given
    "locally". The remaining input values are expected to be satisfied by other
    tasks/imports in the workflow."""

    __slots__ = (
        "_schemas",
        "_repeats",
        "_resources",
        "_inputs",
        "_input_files",
        "_input_file_generator_sources",
        "_output_file_parser_sources",
        "_perturbations",
        "_sequences",
        "_input_sources",
        "_nesting_order",
        "_groups",
        "_name",
        "_defined_input_types",  # assigned in _validate()
    )

    def __init__(
        self,
        schemas: Union[List[TaskSchema], TaskSchema],
        repeats: Optional[Union[int, List[int]]] = 1,
        resources: Optional[Dict[str, Dict]] = None,
        inputs: Optional[List[InputValue]] = None,
        input_files: Optional[List[FileSpec]] = None,
        input_file_generator_sources: Optional[List] = None,
        output_file_parser_sources: Optional[List] = None,
        perturbations: Optional[List[ValuePerturbation]] = None,
        sequences: Optional[List[ValueSequence]] = None,
        input_sources: Optional[Dict[str, InputSource]] = None,
        nesting_order: Optional[Dict] = None,
        groups: Optional[List[ElementGroup]] = None,
    ):
        self._schemas = schemas if isinstance(schemas, list) else [schemas]
        self._repeats = repeats
        self._resources = resources or {
            "main": {}
        }  # TODO: use action names from schemas
        self._inputs = inputs or []
        self._input_files = input_files or []
        self._input_file_generator_sources = input_file_generator_sources or []
        self._output_file_parser_sources = output_file_parser_sources or []
        self._perturbations = perturbations or []
        self._sequences = sequences or []
        self._input_sources = input_sources or {}
        self._nesting_order = nesting_order or {}
        self._groups = GroupList(*(groups or ()))

        print(f"tasktemplate init nesting_order: {nesting_order}")

        self._validate()
        self._name = self._get_name()

    def _validate(self):

        # TODO: check a nesting order specified for each sequence?

        names = set(i.objective.name for i in self.schemas)
        if len(names) > 1:
            raise TaskTemplateMultipleSchemaObjectives(
                f"All task schemas used within a task must have the same "
                f"objective, but found multiple objectives: {list(names)!r}"
            )

        input_types = [
            i.parameter.typ for i in self.get_non_sub_parameter_input_values()
        ]
        dup_params = get_duplicate_items(input_types)
        if dup_params:
            raise TaskTemplateMultipleInputValues(
                f"The following parameters are associated with multiple input value "
                f"definitions: {dup_params!r}."
            )

        unexpected_types = set(input_types) - self.all_schema_input_types
        if unexpected_types:
            raise TaskTemplateUnexpectedInput(
                f"The following input parameters are unexpected: {list(unexpected_types)!r}"
            )

        for k, v in self.nesting_order.items():
            if v < 0:
                raise TaskTemplateInvalidNesting(
                    f"`nesting_order` must be >=0 for all keys, but for key {k!r}, value "
                    f"of {v!r} was specified."
                )

        self._defined_input_types = set(input_types)

    def _get_name(self):
        out = f"{self.objective.name}"
        for idx, schema_i in enumerate(self.schemas, start=1):
            need_and = idx < len(self.schemas) and (
                self.schemas[idx].method or self.schemas[idx].implementation
            )
            out += (
                f"{f'_{schema_i.method}' if schema_i.method else ''}"
                f"{f'_{schema_i.implementation}' if schema_i.implementation else ''}"
                f"{f'_and' if need_and else ''}"
            )
        return out

    @classmethod
    def from_spec(cls, spec, all_schemas, all_parameters):
        key = (
            spec.pop("objective"),
            spec.pop("method", None),
            spec.pop("implementation", None),
        )  # TODO: maybe don't mutate spec?
        spec["schemas"] = all_schemas[
            key
        ]  # TODO parse multiple methods/impl not multiple schemas

        sequences = spec.pop("sequences", [])
        inputs_spec = spec.pop("inputs", [])
        input_sources_spec = spec.pop("input_sources", {})
        perturbs_spec = spec.pop("perturbations", {})
        nesting_order = spec.pop("nesting_order", {})

        for k in list(nesting_order.keys()):
            new_k = tuple(k.split("."))
            nesting_order[new_k] = nesting_order.pop(k)

        print(f"nesting_order: {nesting_order}")

        inputs = []
        if isinstance(inputs_spec, dict):
            # transform to a list:
            for input_path, input_val in inputs_spec.items():
                is_sequence = input_path.endswith("[]")
                if is_sequence:
                    input_path = input_path.split("[]")[0]
                input_path = input_path.split(".")
                inputs.append(
                    {
                        "parameter": input_path[0],
                        "path": input_path[1:],
                        "value": input_val[0] if is_sequence else input_val,
                    }
                )
                if is_sequence:
                    seq_path = ["inputs"] + input_path
                    sequences.append(
                        {
                            "path": seq_path,
                            "values": input_val,
                            "nesting_order": nesting_order.get(tuple(seq_path)),
                        }
                    )
        else:
            inputs = inputs_spec

        # add any nesting orders that are defined in the sequences:
        for i in sequences:
            if "nesting_order" in i:
                nesting_order.update({tuple(i["path"]): i["nesting_order"]})

        perturbs = [{"name": p_name, **p} for p_name, p in perturbs_spec.items()]

        spec.update(
            {
                "inputs": [InputValue.from_spec(i, all_parameters) for i in inputs],
                "sequences": [ValueSequence.from_spec(i) for i in sequences],
                "perturbations": [ValuePerturbation.from_spec(i) for i in perturbs],
                "nesting_order": nesting_order,
                "input_sources": {
                    i: [InputSource.from_spec(j) for j in sources]
                    for i, sources in input_sources_spec.items()
                },
            }
        )

        return cls(**spec)

    @property
    def schemas(self):
        return self._schemas

    @property
    def repeats(self):
        return self._repeats

    @property
    def resources(self):
        return self._resources

    @property
    def inputs(self):
        return self._inputs

    @property
    def input_files(self):
        return self._input_files

    @property
    def input_file_generator_sources(self):
        return self._input_file_generator_sources

    @property
    def output_file_parser_sources(self):
        return self._output_file_parser_sources

    @property
    def perturbations(self):
        return self._perturbations

    @property
    def sequences(self):
        return self._sequences

    @property
    def input_sources(self):
        return self._input_sources

    @property
    def nesting_order(self):
        return self._nesting_order

    @property
    def groups(self):
        return self._groups

    @property
    def name(self):
        return self._name

    @property
    def objective(self):
        return self.schemas[0].objective

    @property
    def all_schema_inputs(self) -> Tuple[SchemaInput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.inputs)

    @property
    def all_schema_outputs(self) -> Tuple[SchemaOutput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.outputs)

    @property
    def all_schema_input_types(self):
        """Get the set of all schema input types (over all specified schemas)."""
        return {inp_j for schema_i in self.schemas for inp_j in schema_i.input_types}

    @property
    def all_schema_output_types(self):
        """Get the set of all schema output types (over all specified schemas)."""
        return {out_j for schema_i in self.schemas for out_j in schema_i.output_types}

    @property
    def universal_input_types(self):
        """Get input types that are associated with all schemas"""

    @property
    def non_universal_input_types(self):
        """Get input types for each schema that are non-universal."""

    @property
    def defined_input_types(self):
        return self._defined_input_types

    @property
    def undefined_input_types(self):
        return self.all_schema_input_types - self.defined_input_types

    @property
    def undefined_inputs(self):
        return [
            inp_j
            for schema_i in self.schemas
            for inp_j in schema_i.inputs
            if inp_j.typ in self.undefined_input_types
        ]

    @property
    def unsourced_inputs(self):
        """Get schema input types for which no input sources are currently specified."""
        return self.all_schema_input_types - set(self.input_sources.keys())

    @property
    def provides_parameters(self):
        return tuple(j for schema in self.schemas for j in schema.provides_parameters)

    def get_sub_parameter_input_values(self):
        return [i for i in self.inputs if i.is_sub_value]

    def get_non_sub_parameter_input_values(self):
        return [i for i in self.inputs if not i.is_sub_value]

    def add_group(
        self, name: str, where: ElementFilter, group_by_distinct: ParameterPath
    ):
        group = ElementGroup(
            name=name, where=where, group_by_distinct=group_by_distinct
        )
        self.groups.add_object(group)

    def get_input_multiplicities(self, missing_multiplicities=None):
        """Get multiplicities for all inputs."""

        if self.undefined_input_types:
            missing_inputs = self.undefined_input_types - set(
                (missing_multiplicities or {}).keys()
            )
            if missing_inputs:
                raise MissingInputs(
                    f"The following inputs are not assigned values, so task input "
                    f"multiplicities cannot be resolved: {list(missing_inputs)!r}."
                )

        input_multiplicities = []
        for i in self.input_values:
            if i.sequences:
                for seq in i.sequences:
                    address = tuple([i.parameter.typ] + seq.address)
                    address_str = ".".join(address)
                    input_multiplicities.append(
                        {
                            "address": address,
                            "multiplicity": len(seq.values),
                            "nesting_order": self.nesting_order[address_str],
                        }
                    )
            else:
                input_multiplicities.append(
                    {
                        "address": (i.parameter.typ,),
                        "multiplicity": 1,
                        "nesting_order": -1,
                    }
                )

        for inp_type, multi in (missing_multiplicities or {}).items():
            input_multiplicities.append(
                {
                    "address": (inp_type,),
                    "multiplicity": multi,
                    "nesting_order": self.nesting_order[inp_type],
                }
            )

        return input_multiplicities


class Task:
    """Class to represent a Task as positioned within a Workflow or WorkflowTemplate."""

    def __init__(
        self,
        template: TaskTemplate,
        workflow: Union[Workflow, WorkflowTemplate],
        initial_index: int,
    ):

        self._template = template
        self._workflow = workflow
        self._initial_index = initial_index

    @property
    def template(self):
        return self._template

    @property
    def workflow(self):
        return self._workflow

    @property
    def element_indices(self):
        return self.workflow.element_indices[self.index]

    @property
    def num_elements(self):
        return len(self.element_indices)

    @property
    def index(self):
        """Zero-based position within the workflow. Uses initial index if appending to the
        workflow is not complete."""
        try:
            return index(self.workflow.tasks, self)
        except ValueError:
            return self._initial_index

    @property
    def unique_name(self):
        name_repeat_index = self.workflow.name_repeat_indices[self.index]
        add_rep = f"{'_' + name_repeat_index if name_repeat_index > 1 else ''}"
        return f"{self.template.name}{add_rep}"



from pathlib import Path
from ruamel.yaml import YAML, safe_load


class Config:

    __conf = {}

    @staticmethod
    def get_config_file(config_dir):

        config_file = config_dir.joinpath("config.yaml")
        with config_file.open() as handle:
            config_dat = safe_load(handle)

        return config_dat, config_file

    @staticmethod
    def resolve_config_dir(config_dir=None):

        if not config_dir:
            config_dir = Path("../hpcflow/")
        else:
            config_dir = Path(config_dir)

        if not config_dir.is_dir():
            print("Configuration directory does not exist. Generating.")
            config_dir.mkdir()

        return config_dir

    @staticmethod
    def set_config(config_dir=None):
        """Load configuration from a YAML file."""

        config_dir = Config.resolve_config_dir(config_dir)
        config_dat, _ = Config.get_config_file(config_dir)

        Config.__conf.update(**config_dat)



from dataclasses import dataclass
from pathlib import Path
from re import I
from typing import Dict, List, Union


from hpcflow.parameters import Parameter
from hpcflow.environment import Environment
from hpcflow.utils import search_dir_files_by_regex


@dataclass
class FileSpec:
    label: str
    name: str

    def __post_init__(self):
        self.name = FileNameSpec(self.name) if isinstance(self.name, str) else self.name
        self.stem = self.name.stem
        self.ext = self.name.ext

    def value(self, directory=None):
        return self.name.value(directory)

    @classmethod
    def from_spec(cls, spec):
        spec["name"] = FileNameSpec.from_spec(spec["name"])
        return cls(**spec)


class FileNameSpec:
    def __init__(self, name, args=None, is_regex=False):
        self.name = name
        self.args = args
        self.is_regex = is_regex
        self.stem = FileNameStem(self)
        self.ext = FileNameExt(self)

    def value(self, directory=None):
        format_args = [i.value(directory) for i in self.args or []]
        value = self.name.format(*format_args)
        if self.is_regex:
            value = search_dir_files_by_regex(value, group=0, directory=directory)
        return value

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    @classmethod
    def from_spec(cls, spec):
        return cls(**spec)


@dataclass
class FileNameStem:
    file_name: FileNameSpec

    def value(self, directory=None):
        return Path(self.file_name.value(directory)).stem


@dataclass
class FileNameExt:
    file_name: FileNameSpec

    def value(self, directory=None):
        return Path(self.file_name.value(directory)).suffix


@dataclass
class InputFileGenerator:
    input_file: FileSpec
    inputs: List[Parameter]
    environment: Environment = None

    @classmethod
    def from_spec(cls, label, info, parameters, cmd_files):
        input_file = [i for i in cmd_files if i.label == label][0]
        inputs = [parameters[typ] for typ in info["from_inputs"]]
        return cls(input_file, inputs)


@dataclass
class OutputFileParser:
    output: Parameter
    output_files: List[FileSpec]
    environment: Environment = None
    inputs: List[str] = None
    options: Dict = None

    @classmethod
    def from_spec(cls, param_typ, info, parameters, cmd_files):
        output = parameters[param_typ]
        output_files = [
            [j for j in cmd_files if j.label == label][0]
            for label in info["from_files"]
        ]
        return cls(output, output_files)


class _FileContentsSpecifier:
    """Class to represent the contents of a file, either via a file-system path or directly."""

    def __init__(
        self, path: Union[Path, str] = None, contents: str = None, extension: str = ""
    ):
        self.path = Path(path) if path is not None else path
        self._contents = contents
        self.extension = extension

        if (path is not None and contents is not None) or (
            path is None and contents is None
        ):
            raise ValueError("Specify exactly one of `path` and `contents`.")

    @property
    def contents(self):
        if self.path is not None:
            with self.path.open("r") as fh:
                contents = fh.read()
        else:
            contents = self._contents
        return contents


class InputFile(_FileContentsSpecifier):
    def __init__(
        self,
        file: FileSpec,
        path: Union[Path, str] = None,
        contents: str = None,
        extension: str = "",
    ):
        super().__init__(path, contents, extension)
        self.file = file


class InputFileGeneratorSource(_FileContentsSpecifier):
    def __init__(
        self,
        generator: InputFileGenerator,
        path: Union[Path, str] = None,
        contents: str = None,
        extension: str = "",
    ):
        super().__init__(path, contents, extension)
        self.generator = generator


class OutputFileParserSource(_FileContentsSpecifier):
    def __init__(
        self,
        parser: OutputFileParser,
        path: Union[Path, str] = None,
        contents: str = None,
        extension: str = "",
    ):
        super().__init__(path, contents, extension)
        self.parser = parser



from dataclasses import dataclass, field
from typing import List, Optional, Union

from hpcflow.actions import Action
from hpcflow.errors import MissingActionsError
from hpcflow.parameters import (
    Parameter,
    ParameterPropagationMode,
    SchemaInput,
    SchemaOutput,
    SchemaParameter,
)

from hpcflow.utils import check_valid_py_identifier


@dataclass
class TaskObjective:
    name: str

    def __post_init__(self):
        self.name = check_valid_py_identifier(self.name)


@dataclass
class TaskSchema:

    # TODO: build comprehensive test suite for TaskSchema
    # - decide how schema inputs and outputs are linked to action inputs and outputs?
    # - should the appearance of parameters in the actions determined schema inputs/outputs
    #   - still need to define e.g. defaults, so makes sense to keep inputs/outputs as schema
    #     parameters, and then verify that all action parameters are taken from schema parameters.
    # - should command files be listed as part of the schema? probably, yes.

    objective: Union[TaskObjective, str]
    actions: List[Action]
    method: Optional[str] = None
    implementation: Optional[str] = None
    inputs: Optional[List[Union[Parameter, SchemaInput]]] = field(
        default_factory=lambda: []
    )
    outputs: Optional[List[Union[Parameter, SchemaOutput]]] = field(
        default_factory=lambda: []
    )

    def __post_init__(self):

        self._validate()

    def _validate(self):

        if isinstance(self.objective, str):
            self.objective = TaskObjective(self.objective)

        if self.method:
            self.method = check_valid_py_identifier(self.method)
        if self.implementation:
            self.implementation = check_valid_py_identifier(self.implementation)

        # coerce Parameters to SchemaInputs
        for idx, i in enumerate(self.inputs):
            if isinstance(i, Parameter):
                self.inputs[idx] = SchemaInput(i)

        # coerce Parameters to SchemaOutputs
        for idx, i in enumerate(self.outputs):
            if isinstance(i, Parameter):
                self.outputs[idx] = SchemaOutput(i)

        if not self.actions:
            raise MissingActionsError("A task schema must define at least one Action.")

    @property
    def input_types(self):
        return tuple(i.typ for i in self.inputs)

    @property
    def output_types(self):
        return tuple(i.typ for i in self.outputs)

    @property
    def provides_parameters(self):
        return tuple(
            i
            for i in self.inputs + self.outputs
            if i.propagation_mode != ParameterPropagationMode.NEVER
        )

    @classmethod
    def from_spec(cls, spec, all_parameters, environments, cmd_files):
        spec["inputs"] = [
            SchemaInput.from_spec(i, all_parameters) for i in spec.get("inputs", [])
        ]
        spec["outputs"] = [
            SchemaOutput.from_spec(i, all_parameters) for i in spec.get("outputs", [])
        ]
        spec["actions"] = [
            Action.from_spec(i, environments, all_parameters, cmd_files)
            for i in spec.get("actions")
        ]
        return cls(**spec)

    def get_parameter_dependence(self, parameter: SchemaParameter):
        """Find if/where a given parameter is used by the schema's actions."""
        out = {"input_file_writers": [], "commands": []}
        for act_idx, action in enumerate(self.actions):
            deps = action.get_parameter_dependence(parameter)
            for key in out:
                out[key].extend((act_idx, i) for i in deps[key])
        return out



from dataclasses import dataclass, field
import enum
from typing import Any, List, Optional, Sequence, Union
import numpy as np
from hpcflow.element import ElementFilter
from hpcflow.errors import InputSourceValidationError

from hpcflow.typing_stubs import (
    SubParameter,
    InputValue,
    TaskTemplate,
    TaskSchema,
    WorkflowTemplate,
)
from hpcflow.utils import check_valid_py_identifier


Address = List[Union[int, float, str]]
Numeric = Union[int, float, np.number]


class ParameterPropagationMode(enum.Enum):

    IMPLICIT = 0
    EXPLICIT = 1
    NEVER = 2


@dataclass
class ParameterPath:

    path: Sequence[Union[str, int, float]]
    task: Optional[Union[TaskTemplate, TaskSchema]] = None  # default is "current" task


@dataclass
class Parameter:
    typ: str
    is_file: bool = False
    sub_parameters: List[SubParameter] = field(default_factory=lambda: [])

    def __post_init__(self):
        self.typ = check_valid_py_identifier(self.typ)

    @classmethod
    def from_spec(cls, spec):
        spec["typ"] = spec.pop("type")
        return cls(**spec)


@dataclass
class SubParameter:
    address: Address
    parameter: Parameter


@dataclass
class SchemaParameter:
    @classmethod
    def from_spec(cls, spec, parameters):
        spec["parameter"] = parameters[spec["parameter"]]
        prop_mode = spec.get("propagation_mode")
        if prop_mode:
            spec["propagation_mode"] = getattr(
                ParameterPropagationMode, prop_mode.upper()
            )
        return cls(**spec)

    @property
    def name(self):
        return self.parameter.name

    @property
    def typ(self):
        return self.parameter.typ


@dataclass
class SchemaInput(SchemaParameter):
    """A Parameter as used within a particular schema, for which a default value may be
    applied."""

    parameter: Parameter
    default_value: Optional[InputValue] = None
    propagation_mode: ParameterPropagationMode = ParameterPropagationMode.IMPLICIT

    # can we define elements groups on local inputs as well, or should these be just for
    # elements from other tasks?
    group: Optional[str] = None
    where: Optional[ElementFilter] = None

    def __post_init__(self):
        self._validate()

    def _validate(self):
        if self.default_value is not None:
            if self.default_value.parameter != self.parameter:
                raise ValueError(
                    f"{self.__class__.__name__} `default_value` must be an `InputValue` for "
                    f"parameter: {self.parameter!r}, but specified `InputValue` parameter "
                    f"is: {self.default_value.parameter!r}."
                )

    @property
    def input_or_output(self):
        return "input"

    @classmethod
    def from_spec(cls, spec, parameters):
        if "default_value" in spec:
            spec["default_value"] = InputValue(
                parameter=parameters[spec["parameter"]], value=spec["default_value"]
            )
        return super().from_spec(spec, parameters)


@dataclass
class SchemaOutput(SchemaParameter):
    """A Parameter as outputted from particular task."""

    parameter: Parameter
    propagation_mode: ParameterPropagationMode = ParameterPropagationMode.IMPLICIT

    @property
    def input_or_output(self):
        return "output"


@dataclass
class BuiltinSchemaParameter:
    # builtin inputs (resources,parameter_perturbations,method,implementation
    # builtin outputs (time, memory use, node/hostname etc)
    # - builtin parameters do not propagate to other tasks (since all tasks define the same
    #   builtin parameters).
    # - however, builtin parameters can be accessed if a downstream task schema specifically
    #   asks for them (e.g. for calculating/plotting a convergence test)
    pass


@dataclass
class ValueSequence:
    path: Sequence[Union[str, int, float]]
    values: List[Any]
    nesting_order: int

    def __post_init__(self):
        self.path = tuple(self.path)

    def check_address_exists(self, value):
        """Check a given nested dict/list "address" resolves to somewhere within
        `value`."""
        if self.address:
            sub_val = value
            for i in self.address:
                try:
                    sub_val = sub_val[i]
                except (IndexError, KeyError, TypeError):
                    msg = (
                        f"Address {self.address} does not exist in the base "
                        f"value: {value}"
                    )
                    raise ValueError(msg)

    @classmethod
    def from_spec(cls, spec):
        return cls(**spec)

    @classmethod
    def from_linear_space(cls, start, stop, num=50, address=None, **kwargs):
        values = list(np.linspace(start, stop, num=num, **kwargs))
        return cls(values, address=address)

    @classmethod
    def from_range(cls, start, stop, step=1, address=None):
        if isinstance(step, int):
            return cls(values=list(np.arange(start, stop, step)), address=address)
        else:
            # Use linspace for non-integer step, as recommended by Numpy:
            return cls.from_linear_space(
                start,
                stop,
                num=int((stop - start) / step),
                address=address,
                endpoint=False,
            )


@dataclass
class AbstractInputValue:
    """Class to represent all sequence-able inputs to a task."""


@dataclass
class ResourceSpec(AbstractInputValue):
    pass


@dataclass
class ValuePerturbation(AbstractInputValue):
    name: str
    path: Optional[Sequence[Union[str, int, float]]] = None
    multiplicative_factor: Optional[Numeric] = 1
    additive_factor: Optional[Numeric] = 0

    @classmethod
    def from_spec(cls, spec):
        return cls(**spec)


@dataclass
class InputValue(AbstractInputValue):
    parameter: Union[Parameter, SchemaInput]
    path: Optional[Sequence[Union[str, int, float]]] = None
    value: Optional[Any] = None

    def __post_init__(self):
        self._validate()

    @property
    def is_sub_value(self):
        """True if the value is for a sub part of the parameter (i.e. if `path` is set).
        Sub-values are not added to the base parameter data, but are interpreted as
        single-value sequences."""
        return True if self.path else False

    def _validate(self):
        pass
        # if self.sequences:
        #     seen_addresses = []
        #     for i in self.sequences:
        #         if i.address not in seen_addresses:
        #             seen_addresses.append(i.address)
        #         else:
        #             raise InputValueDuplicateSequenceAddress(
        #                 f"`{i.__class__.__name__}` defined with address {i.address!r} multiple "
        #                 f"times."
        #             )
        #         i.check_address_exists(self.value)

    @classmethod
    def from_spec(cls, spec, all_parameters):
        spec["parameter"] = all_parameters[spec.pop("parameter")]
        return cls(**spec)


@dataclass
class InputSource:
    source: str
    where: Optional[ElementFilter] = None

    # TODO: validate and split?

    def __post_init__(self):
        self._validate()

    def _validate(self):

        parts = self.source.lower().split(".")
        source_type = parts[0]

        allowed = ["imports", "tasks", "local", "default"]
        if source_type not in allowed:
            raise ValueError(
                f"InputSource `source` specified as {source_type!r}, but must be one "
                f"of: {allowed!r}."
            )

        if source_type == "tasks":
            task_ref = parts[1]
            task_source_type = parts[2]
            allowed_task_types = ["inputs", "outputs"]
            if task_source_type not in allowed_task_types:
                raise ValueError(
                    f"InputSource `source` with source type 'tasks' must be in the format: "
                    f"'tasks.[task_reference].inputs' or 'tasks.[task_reference].outputs', "
                    f"where [task_reference] is the unique name of a workflow task, specified "
                    f"source was: {self.source!r}."
                )
            self._task_ref = task_ref
            self._task_source_type = task_source_type

        elif source_type == "imports":
            self._imports_ref = parts[1]

        if (source_type == "local" and len(parts) > 1) or (
            source_type == "imports" and len(parts) > 2
        ):
            raise ValueError(f"InputSource source not understood: {self.source!r}.")

        if source_type in ["local", "default"] and self.where is not None:
            raise ValueError(
                f"Element filter via the `where` argument is not supported for InputSource "
                f"with source type {source_type!r}."
            )

        self._source_type = source_type

    @property
    def source_type(self):
        return self._source_type

    @property
    def task_ref(self):
        return self._task_ref

    @property
    def task_source_type(self):
        return self._task_source_type

    @property
    def imports_ref(self):
        return self._imports_ref

    @classmethod
    def from_spec(cls, spec):
        if "where" in spec:
            spec["where"] = ElementFilter.from_spec(spec["where"])
        return cls(**spec)

    def validate(
        self,
        schema_input: SchemaInput,
        task_template: TaskTemplate,
        workflow_like: WorkflowTemplate,
    ):
        """Check a supplied source is valid for a given workflow (template)."""

        if self.source_type == "tasks":

            # check referenced task exists:
            try:
                ref_task = getattr(workflow_like.tasks, self.task_ref)
            except AttributeError:
                raise InputSourceValidationError(
                    f"InputSource {self.source!r} cannot be resolved within the workflow."
                )

            # check the parameter is provided by the referenced task:
            if self.task_source_type == "inputs":
                if schema_input.typ not in ref_task.template.all_schema_input_types:
                    raise InputSourceValidationError(
                        f"Input parameter {schema_input.typ!r} cannot be sourced from task "
                        f"{self.task_ref!r}, since the task schemas do not provide this "
                        f"parameter as an input. Available inputs from this task are: "
                        f"{ref_task.template.all_schema_input_types!r}."
                    )
            elif self.task_source_type == "outputs":
                if schema_input.typ not in ref_task.template.all_schema_output_types:
                    raise InputSourceValidationError(
                        f"Input parameter {schema_input.typ!r} cannot be sourced from task "
                        f"{self.task_ref!r}, since the task schemas do not provide this "
                        f"parameter as an output. Available outputs from this task are: "
                        f"{ref_task.template.all_schema_output_types!r}."
                    )

        elif self.source_type == "imports":
            raise NotImplementedError("ain't dunit yet")

        elif self.source_type == "default":
            # check a default value exists:
            if schema_input.default_value is None:
                raise InputSourceValidationError(
                    f"Input parameter {schema_input.typ!r} cannot be sourced from the "
                    f"default value, because no default value is specified."
                )

        elif self.source_type == "local":
            # check local value is supplied by schema:
            if schema_input.typ not in task_template.all_schema_input_types:
                raise InputSourceValidationError(
                    f"Input parameter {schema_input.typ!r} cannot be sourced from the "
                    f"local inputs, because the schema does not provide such an input. "
                    f"Available inputs defined by the schema are: "
                    f"{task_template.all_schema_inputs!r}."
                )

            # check local value exists:
            if schema_input.typ not in task_template.defined_input_types:
                raise InputSourceValidationError(
                    f"Input parameter {schema_input.typ!r} cannot be sourced from the "
                    f"local inputs, because no local value is defined."
                )



from dataclasses import dataclass, field
import re
from typing import List, Any, Dict, Optional, Sequence, Union

from textwrap import dedent

from hpcflow.errors import DuplicateExecutableError
from hpcflow.utils import check_valid_py_identifier, get_duplicate_items
from hpcflow.object_list import ExecutablesList


@dataclass
class NumCores:
    start: int
    stop: int
    step: int = None

    def __post_init__(self):
        if self.step is None:
            self.step = 1

    def __contains__(self, x):
        if x in range(self.start, self.stop + 1, self.step):
            return True
        else:
            return False

    def __eq__(self, other):
        if (
            type(self) == type(other)
            and self.start == other.start
            and self.stop == other.stop
            and self.step == other.step
        ):
            return True
        return False


@dataclass
class ExecutableInstance:
    parallel_mode: str
    num_cores: Any
    command: str

    def __post_init__(self):
        if not isinstance(self.num_cores, dict):
            self.num_cores = {"start": self.num_cores, "stop": self.num_cores}
        if not isinstance(self.num_cores, NumCores):
            self.num_cores = NumCores(**self.num_cores)

    def __eq__(self, other):
        if (
            type(self) == type(other)
            and self.parallel_mode == other.parallel_mode
            and self.num_cores == other.num_cores
            and self.command == other.command
        ):
            return True
        return False

    @classmethod
    def from_spec(cls, spec):
        return cls(**spec)


@dataclass
class Executable:
    label: str
    instances: List[ExecutableInstance] = field(repr=False, default_factory=lambda: [])
    environment: Any = field(default=None, repr=False)

    def __post_init__(self):
        self.label = check_valid_py_identifier(self.label)

    def __eq__(self, other):
        if (
            type(self) == type(other)
            and self.label == other.label
            and self.instances == other.instances
            and self.environment.name == other.environment.name
        ):
            return True
        return False

    def filter_instances(self, parallel_mode=None, num_cores=None):
        out = []
        for i in self.instances:
            if parallel_mode is None or i.parallel_mode == parallel_mode:
                if num_cores is None or num_cores in i.num_cores:
                    out.append(i)
        return out

    @classmethod
    def from_spec(cls, spec):
        spec["instances"] = [
            ExecutableInstance.from_spec(i) for i in spec.get("instances", [])
        ]
        return cls(**spec)


@dataclass
class Environment:
    name: str
    setup: Optional[Sequence] = None
    executables: Optional[List[Executable]] = field(default_factory=lambda: [])

    def __post_init__(self):
        for i in self.executables:
            i.environment = self

        if self.setup:
            if isinstance(self.setup, str):
                self.setup = tuple(
                    i.strip() for i in dedent(self.setup).strip().split("\n")
                )
            elif not isinstance(self.setup, tuple):
                self.setup = tuple(self.setup)

        self.executables = ExecutablesList(*self.executables)

        self._validate()

    def __eq__(self, other):
        if (
            type(self) == type(other)
            and self.setup == other.setup
            and self.executables == other.executables
        ):
            return True
        return False

    def _validate(self):
        dup_labels = get_duplicate_items(i.label for i in self.executables)
        if dup_labels:
            raise DuplicateExecutableError(
                f"Executables must have unique `label`s within each environment, but "
                f"found label(s) multiple times: {dup_labels!r}"
            )

    @classmethod
    def from_spec(cls, spec):
        spec["executables"] = [
            Executable.from_spec(i) for i in spec.get("executables", [])
        ]
        return cls(**spec)



import click
from hpcflow import __version__


@click.group(name="hpcflow")
@click.version_option(version=__version__, package_name="hpcflow", prog_name="hpcflow")
def cli():
    pass


@cli.command()
def make_workflow():
    """Command on hpcflow"""


if __name__ == "__main__":
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hpcflow', 'hpcflow.data']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.4,<9.0.0',
 'ruamel.yaml>=0.17.20,<0.18.0',
 'valida>=0.2.0,<0.3.0',
 'zarr>=2.10.3,<3.0.0']

extras_require = \
{'pyinstaller': ['pyinstaller>=4.10,<5.0']}

entry_points = \
{'console_scripts': ['hpcflow = hpcflow.cli:cli']}

setup_kwargs = {
    'name': 'hpcflow-new',
    'version': '0.80.0a2',
    'description': 'Computational workflow management',
    'long_description': '<img src="docs/source/_static/images/logo-v2.png" width="250" alt="hpcFlow logo"/>\n\n**hpcFlow manages your scientific workflows**\n\nDocumentation: [https://hpcflow.github.io/docs](https://hpcflow.github.io/docs)\n\n## Acknowledgements\n\n\nhpcFlow was developed using funding from the [LightForm](https://lightform.org.uk/) EPSRC programme grant ([EP/R001715/1](https://gow.epsrc.ukri.org/NGBOViewGrant.aspx?GrantRef=EP/R001715/1))\n\n<img src="https://lightform-group.github.io/wiki/assets/images/site/lightform-logo.png" width="150"/>\n',
    'author': 'aplowman',
    'author_email': 'adam.plowman@manchester.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from multiprocessing.sharedctypes import Value


class InputValueDuplicateSequenceAddress(ValueError):
    pass


class TaskTemplateMultipleSchemaObjectives(ValueError):
    pass


class TaskTemplateUnexpectedInput(ValueError):
    pass


class TaskTemplateMultipleInputValues(ValueError):
    pass


class InvalidIdentifier(ValueError):
    pass


class MissingInputs(Exception):
    pass


class TaskTemplateInvalidNesting(ValueError):
    pass


class TaskSchemaSpecValidationError(Exception):
    pass


class WorkflowSpecValidationError(Exception):
    pass


class InputSourceValidationError(Exception):
    pass


class EnvironmentSpecValidationError(Exception):
    pass


class DuplicateExecutableError(ValueError):
    pass


class MissingActionsError(ValueError):
    pass


class MissingCompatibleActionEnvironment(Exception):
    pass


class MissingActionEnvironment(Exception):
    pass
